"""Tests juju-core environment watchers."""

import unittest

import mock

from deployer.env import watchers
from deployer.utils import ErrorExit


class TestWaitForUnits(unittest.TestCase):

    def make_watch(self, *changesets):
        """Create and return a mock watcher returning the given change sets."""
        watch = mock.MagicMock()
        watch.__iter__.return_value = changesets
        return watch

    def make_change(self, unit_name, status):
        """Create and return a juju-core mega-watcher unit change."""
        service = unit_name.split('/')[0]
        return {'Name': unit_name, 'Service': service, 'Status': status}

    def test_success(self):
        # It is possible to watch for units to be started.
        watch = self.make_watch(
            [('unit', 'change', self.make_change('django/0', 'pending')),
             ('unit', 'change', self.make_change('haproxy/1', 'pending'))],
            [('unit', 'change', self.make_change('django/0', 'started')),
             ('unit', 'change', self.make_change('haproxy/1', 'started'))],
        )
        watcher = watchers.WaitForUnits(watch)
        callback = mock.Mock()
        watcher.run(callback)
        # The callback has been called once for each change in the change sets,
        # excluding the initial state.
        self.assertEqual(2, callback.call_count)
        # The watcher has been properly stopped.
        watch.stop.assert_called_once_with()

    def test_watch_restart(self):
        # It is possible for a watch to be restarted.
        watch = self.make_watch(
            [('unit', 'change', self.make_change('django/0', 'pending')),
             ('unit', 'change', self.make_change('haproxy/1', 'pending'))],
            [('unit', 'change', self.make_change('django/0', 'started'))],
            # On a restart we get reinformed of things already current
            [('unit', 'change', self.make_change('django/0', 'started')),
             ('unit', 'change', self.make_change('haproxy/1', 'pending'))],
            [('unit', 'change', self.make_change('django/0', 'started')),
             ('unit', 'change', self.make_change('haproxy/1', 'started'))]
        )
        watcher = watchers.WaitForUnits(watch)
        callback = mock.Mock()
        watcher.run(callback)
        # The callback has been called once for each change in the change sets,
        # excluding the initial state.
        self.assertEqual(5, callback.call_count)
        # The watcher has been properly stopped.
        watch.stop.assert_called_once_with()

    def test_errors_handling(self):
        # Errors can be handled providing an on_errors callable.
        watch = self.make_watch(
            [('unit', 'change', self.make_change('django/0', 'pending')),
             ('unit', 'change', self.make_change('django/42', 'pending')),
             ('unit', 'change', self.make_change('haproxy/1', 'pending'))],
            [('unit', 'change', self.make_change('django/0', 'pending')),
             ('unit', 'change', self.make_change('django/42', 'error')),
             ('unit', 'change', self.make_change('haproxy/1', 'error'))],
            [('unit', 'change', self.make_change('django/0', 'error')),
             ('unit', 'change', self.make_change('django/42', 'error')),
             ('unit', 'change', self.make_change('haproxy/1', 'error'))],
        )
        on_errors = mock.Mock()
        watcher = watchers.WaitForUnits(watch, on_errors=on_errors)
        watcher.run()
        # The watcher has been properly stopped.
        watch.stop.assert_called_once_with()
        # The errors handler has been called once for each changeset containing
        # errors.
        self.assertEqual(2, on_errors.call_count)
        on_errors.assert_has_calls([
            mock.call([
                {'Status': 'error', 'Name': 'django/42', 'Service': 'django'},
                {'Status': 'error', 'Name': 'haproxy/1', 'Service': 'haproxy'}
            ]),
            mock.call([
                {'Status': 'error', 'Name': 'django/0', 'Service': 'django'}]),
        ])

    def test_specific_services(self):
        # It is possible to only watch units belonging to specific services.
        watch = self.make_watch(
            [('unit', 'change', self.make_change('django/0', 'pending')),
             ('unit', 'change', self.make_change('django/42', 'pending')),
             ('unit', 'change', self.make_change('haproxy/1', 'pending')),
             ('unit', 'change', self.make_change('haproxy/47', 'pending'))],
            [('unit', 'change', self.make_change('django/0', 'pending')),
             ('unit', 'change', self.make_change('django/42', 'started')),
             ('unit', 'change', self.make_change('haproxy/1', 'error')),
             ('unit', 'change', self.make_change('haproxy/47', 'pending'))],
            [('unit', 'change', self.make_change('django/0', 'error')),
             ('unit', 'change', self.make_change('django/42', 'started')),
             ('unit', 'change', self.make_change('haproxy/1', 'error')),
             ('unit', 'change', self.make_change('haproxy/47', 'pending'))],
        )
        on_errors = mock.Mock()
        watcher = watchers.WaitForUnits(
            watch, services=['django'], on_errors=on_errors)
        watcher.run()
        # The watcher has been properly stopped, even if haproxy/47 is pending.
        watch.stop.assert_called_once_with()
        # The errors handler has been called for the django error, not for the
        # haproxy one.
        on_errors.assert_called_once_with(
            [{'Status': 'error', 'Name': 'django/0', 'Service': 'django'}])

    def test_goal_states(self):
        # It is possible to watch for the given goal state
        watch = self.make_watch(
            [('unit', 'change', self.make_change('django/0', 'pending')),
             ('unit', 'change', self.make_change('haproxy/1', 'pending'))],
        )
        watcher = watchers.WaitForUnits(watch, goal_state='pending')
        callback = mock.Mock()
        watcher.run(callback)
        # Since all the units are already pending, the watcher has been
        # properly stopped.
        watch.stop.assert_called_once_with()


class TestLogOnErrors(unittest.TestCase):

    def setUp(self):
        # Set up a mock environment.
        self.env = mock.Mock()

    def test_returned_callable(self):
        # The returned function uses the env to log errors.
        callback = watchers.log_on_errors(self.env)
        self.assertEqual(self.env.log_errors, callback)


class TestExitOnErrors(unittest.TestCase):

    def setUp(self):
        # Set up a mock environment.
        self.env = mock.Mock()

    def test_returned_callable(self):
        # The returned function uses the env to log errors and then exits the
        # application.
        callback = watchers.exit_on_errors(self.env)
        with self.assertRaises(ErrorExit):
            callback('bad wolf')
        self.env.log_errors.assert_called_once_with('bad wolf')


class TestRaiseOnErrors(unittest.TestCase):

    def test_returned_callable(self):
        # The returned function raises the given exception passing the errors.
        callback = watchers.raise_on_errors(ValueError)
        with self.assertRaises(ValueError) as cm:
            callback('bad wolf')
        self.assertEqual('bad wolf', bytes(cm.exception))



import StringIO

from .base import Base
from deployer.env import watchers
from deployer.env.py import PyEnvironment
from deployer.errors import UnitErrors
from deployer.utils import setup_logging, ErrorExit


class FakePyEnvironment(PyEnvironment):

    def __init__(self, name, status):
        super(FakePyEnvironment, self).__init__(name)
        self._status = status

    def status(self):
        return self._status


class PyEnvironmentTest(Base):

    def setUp(self):
        self.output = setup_logging(
            debug=True, verbose=True, stream=StringIO.StringIO())

    def test_wait_for_units_error_no_exit(self):
        env = FakePyEnvironment(
            "foo",
            {"services":
             {"wordpress":
              {"units":
               {"wordpress/0":
                {"agent-state": "install-error",
                 "machine": 1},
                },
               },
              },
             })
        with self.assertRaises(UnitErrors) as cm:
            env.wait_for_units(
                timeout=240, on_errors=watchers.raise_on_errors(UnitErrors))
        errors = cm.exception.errors
        self.assertEqual(1, len(errors))
        unit = errors[0]
        self.assertEqual("wordpress/0", unit["name"])
        self.assertEqual("install-error", unit["agent-state"])
        self.assertEqual(1, unit["machine"])

    def test_wait_for_units_error_exit(self):
        env = FakePyEnvironment(
            "foo",
            {"services":
             {"wordpress":
              {"units":
               {"wordpress/0":
                {"agent-state": "install-error",
                 "machine": 1},
                },
               },
              },
             })
        with self.assertRaises(ErrorExit):
            env.wait_for_units(
                timeout=240, on_errors=watchers.exit_on_errors(env))
        output = self.output.getvalue()
        self.assertIn('The following units had errors:', output)
        self.assertIn(
            'unit: wordpress/0: machine: 1 agent-state: install-error', output)

    def test_wait_for_units_sub_error_no_exit(self):
        env = FakePyEnvironment(
            "foo",
            {"services":
             {"wordpress":
              {"units":
               {"wordpress/0":
                {"agent-state": "started",
                 "machine": 1,
                 "subordinates":
                 {"nrpe/0":
                  {"agent-state": "install-error"},
                  }
                 },
                },
               },
              },
             })
        with self.assertRaises(UnitErrors) as cm:
            env.wait_for_units(
                timeout=240, on_errors=watchers.raise_on_errors(UnitErrors))
        errors = cm.exception.errors
        self.assertEqual(1, len(errors))
        unit = errors[0]
        self.assertEqual("nrpe/0", unit["name"])
        self.assertEqual("install-error", unit["agent-state"])
        self.assertEqual(1, unit["machine"])

    def test_wait_for_units_no_error_no_exit(self):
        env = FakePyEnvironment(
            "foo",
            {"services":
             {"wordpress":
              {"units":
               {"wordpress/0":
                {"agent-state": "started",
                 "machine": 1,
                 "subordinates":
                 {"nrpe/0":
                  {"agent-state": "started"},
                  }
                 },
                },
               },
              },
             })
        try:
            env.wait_for_units(
                timeout=240, on_errors=watchers.raise_on_errors(UnitErrors))
        except UnitErrors as err:
            self.fail("Unexpected exception: %s" % err)

    def test_wait_for_units_relation_error_no_exit(self):
        env = FakePyEnvironment(
            "foo",
            {"services":
             {"wordpress":
              {"units":
               {"wordpress/0":
                {"agent-state": "started",
                 "machine": 1,
                 "relation-errors":
                 {"nrpe": ["nrpe/1"]},
                 },
                },
               },
              },
             })
        with self.assertRaises(UnitErrors) as cm:
            env.wait_for_units(
                timeout=240, on_errors=watchers.raise_on_errors(UnitErrors))
        errors = cm.exception.errors
        self.assertEqual(1, len(errors))
        unit = errors[0]
        self.assertEqual("wordpress/0", unit["name"])
        self.assertEqual("relation-error: nrpe", unit["agent-state"])
        self.assertEqual(1, unit["machine"])

    def test_wait_for_units_subordinate_no_unit_no_error_no_exit(self):
        env = FakePyEnvironment(
            "foo",
            {"services":
             {"nrpe":
              {"subordinate": "true",
               "units": {},
               },
              },
             })
        try:
            env.wait_for_units(
                timeout=240, on_errors=watchers.raise_on_errors(UnitErrors))
        except UnitErrors as err:
            self.fail("Unexpected exception: %s" % err)



import os

import mock

from deployer.config import ConfigStack
from deployer.action.importer import Importer
from deployer.utils import yaml_dump, yaml_load

from base import (
    Base,
    patch_env_status,
    skip_if_offline,
)


class Options(dict):

    def __getattr__(self, key):
        return self[key]


class ImporterTest(Base):

    def setUp(self):
        self.juju_home = self.mkdir()
        self.change_environment(JUJU_HOME=self.juju_home)
        self.options = Options({
            'bootstrap': False,
            'branch_only': False,
            'configs': [os.path.join(self.test_data_dir, 'wiki.yaml')],
            'debug': True,
            'deploy_delay': 0,
            'destroy_services': None,
            'diff': False,
            'find_service': None,
            'ignore_errors': False,
            'list_deploys': False,
            'no_local_mods': True,
            'no_relations': False,
            'overrides': None,
            'rel_wait': 60,
            'retry_count': 0,
            'series': None,
            'skip_unit_wait': False,
            'terminate_machines': False,
            'timeout': 2700,
            'update_charms': False,
            'verbose': True,
            'watch': False})

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer(self, mock_time):
        # Trying to track down where this comes from http://pad.lv/1243827
        stack = ConfigStack(self.options.configs)
        deploy = stack.get('wiki')
        env = mock.MagicMock()
        patch_env_status(env, {'wiki': 1, 'db': 1})
        importer = Importer(env, deploy, self.options)
        importer.run()

        config = {'name': '$hi_world _are_you_there? {guess_who}'}
        self.assertEqual(
            env.method_calls[3], mock.call.deploy(
                'wiki', 'cs:precise/mediawiki', '', config, None, 1, None))
        env.add_relation.assert_called_once_with('wiki', 'db')

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_no_relations(self, mock_time):
        self.options.no_relations = True
        stack = ConfigStack(self.options.configs)
        deploy = stack.get('wiki')
        env = mock.MagicMock()
        patch_env_status(env, {'wiki': 1, 'db': 1})
        importer = Importer(env, deploy, self.options)
        importer.run()
        self.assertFalse(env.add_relation.called)

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_add_machine_series(self, mock_time):
        self.options.configs = [
            os.path.join(self.test_data_dir, 'v4', 'series.yaml')]
        stack = ConfigStack(self.options.configs)
        deploy = stack.get(self.options.configs[0])
        env = mock.MagicMock()
        patch_env_status(env, {'mediawiki': 1, 'mysql': 1})
        importer = Importer(env, deploy, self.options)
        importer.run()

        self.assertEqual(env.add_machine.call_count, 2)
        self.assertEqual(
            env.add_machine.call_args_list[0][1],
            {'series': 'precise', 'constraints': 'mem=512M'})
        self.assertEqual(
            env.add_machine.call_args_list[1][1],
            {'series': 'trusty', 'constraints': 'mem=512M'})

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_existing_machine(self, mock_time):
        self.options.configs = [
            os.path.join(self.test_data_dir, 'v4',
                         'container-existing-machine.yaml')]
        stack = ConfigStack(self.options.configs)
        deploy = stack.get(self.options.configs[0])
        env = mock.MagicMock()
        patch_env_status(env, {'mediawiki': 1, 'mysql': 1}, machines=[1])

        importer = Importer(env, deploy, self.options)
        importer.run()
        self.assertFalse(env.add_machine.called)



""" Unittest for juju-deployer diff action (--diff) """
# pylint: disable=C0103
import StringIO
import os
import shutil
import tempfile
import unittest

from deployer.config import ConfigStack
from deployer.env.mem import MemoryEnvironment
from deployer.utils import setup_logging

from .base import Base, skip_if_offline, TEST_OFFLINE, TEST_OFFLINE_REASON
from ..action.diff import Diff


@skip_if_offline
class DiffTest(Base):

    def setUp(self):
        self.output = setup_logging(
            debug=True, verbose=True, stream=StringIO.StringIO())

    # Because fetch_charms is expensive, do it once for all tests
    @classmethod
    def setUpClass(cls):
        super(DiffTest, cls).setUpClass()
        # setUpClass not being skipped, here, could have to do with
        # decorator on derived class.  So skip explicitly.
        if TEST_OFFLINE:
            raise unittest.SkipTest(TEST_OFFLINE_REASON)
        deployment = ConfigStack(
            [os.path.join(
                cls.test_data_dir, "blog.yaml")]).get('wordpress-prod')
        cls._dir = tempfile.mkdtemp()
        os.mkdir(os.path.join(cls._dir, "precise"))
        deployment.repo_path = cls._dir
        deployment.fetch_charms()
        deployment.resolve()
        cls._deployment = deployment

    @classmethod
    def tearDownClass(cls):
        super(DiffTest, cls).tearDownClass()
        # TearDownClass not being skipped, here, could have to do with
        # decorator on derived class.  So skip explicitly.
        if TEST_OFFLINE:
            raise unittest.SkipTest(TEST_OFFLINE_REASON)
        shutil.rmtree(cls._dir)

    @classmethod
    def get_deployment(cls):
        """ Return saved deployment at class initialization
        """
        return cls._deployment

    def test_diff_nil(self):
        dpl = self.get_deployment()
        # No changes, assert nil diff
        env = MemoryEnvironment(dpl.name, dpl)
        diff = Diff(env, dpl, {}).do_diff()
        self.assertEqual(diff, {})

    def test_diff_num_units(self):
        # Removing 1 unit must show -1 'num_units'
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.remove_unit(env.status()['services']['haproxy']['units'][0])
        diff = Diff(env, dpl, {}).do_diff()
        self.assertEqual(
            diff['services']['modified']['haproxy']['num_units'], -1)
        # re-adding a unit -> nil diff
        env.add_units('haproxy', 1)
        diff = Diff(env, dpl, {}).do_diff()
        self.assertEqual(diff, {})

    def test_diff_config(self):
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.set_config('blog', {'tuning': 'bare'})
        diff = Diff(env, dpl, {}).do_diff()
        mod_blog = diff['services']['modified']['blog']
        self.assertTrue(mod_blog['env-config']['tuning'] !=
                        mod_blog['cfg-config']['tuning'])
        self.assertEquals(mod_blog['env-config']['tuning'], 'bare')

    def test_diff_config_many(self):
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.set_config('blog', {'tuning': 'bare', 'engine': 'duck'})
        diff = Diff(env, dpl, {}).do_diff()
        mod_blog = diff['services']['modified']['blog']
        self.assertEqual(
            set(mod_blog['env-config'].keys()),
            set(['tuning', 'engine']))
        self.assertTrue(mod_blog['env-config']['tuning'] !=
                        mod_blog['cfg-config']['tuning'])
        self.assertTrue(mod_blog['env-config']['engine'] !=
                        mod_blog['cfg-config']['engine'])

    def test_diff_constraints(self):
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.set_constraints('haproxy', 'foo=bar')
        diff = Diff(env, dpl, {}).do_diff()
        mod_haproxy = diff['services']['modified']['haproxy']
        self.assertTrue(
            mod_haproxy['env-constraints'] != mod_haproxy['cfg-constraints'])
        self.assertEqual(mod_haproxy['env-constraints'], {'foo': 'bar'})

    def test_diff_service_destroy(self):
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.destroy_service('haproxy')
        diff = Diff(env, dpl, {}).do_diff()
        self.assertTrue(str(diff['relations']['missing'][0]).find('haproxy')
                        != -1)
        self.assertTrue(diff['services']['missing'].keys() == ['haproxy'])



import base64
import StringIO
import os

from deployer.deployment import Deployment
from deployer.utils import setup_logging, ErrorExit

from .base import Base, skip_if_offline


class FauxService(object):
    """A fake service with a unit_placement attribute, used for testing
    the sort functionality.
    """

    def __init__(self, name=None, unit_placement=None):
        self.name = name
        self.unit_placement = unit_placement


class DeploymentTest(Base):

    def setUp(self):
        self.output = setup_logging(
            debug=True, verbose=True, stream=StringIO.StringIO())

    def get_named_deployment_and_fetch_v3(self, file_name, stack_name):
        deployment = self.get_named_deployment_v3(file_name, stack_name)
        # Fetch charms in order to allow proper late binding config and
        # placement validation.
        repo_path = self.mkdir()
        os.mkdir(os.path.join(repo_path, "precise"))
        deployment.repo_path = repo_path
        deployment.fetch_charms()
        return deployment

    def get_deployment_and_fetch_v4(self, file_name):
        deployment = self.get_deployment_v4(file_name)
        # Fetch charms in order to allow proper late binding config and
        # placement validation.
        repo_path = self.mkdir()
        os.mkdir(os.path.join(repo_path, "precise"))
        deployment.repo_path = repo_path
        deployment.fetch_charms()
        return deployment

    @skip_if_offline
    def test_deployer(self):
        d = self.get_named_deployment_and_fetch_v3('blog.yaml', 'wordpress-prod')
        services = d.get_services()
        self.assertTrue([s for s in services if s.name == "newrelic"])

        # Ensure inheritance order reflects reality, instead of merge value.
        self.assertEqual(
            d.data['inherits'],
            ['wordpress-stage', 'wordpress-base', 'metrics-base'])

        # Load up overrides and resolves
        d.load_overrides(["key=abc"])
        d.resolve()

        # Verify include-base64
        self.assertEqual(d.get_service('newrelic').config, {'key': 'abc'})
        self.assertEqual(
            base64.b64decode(d.get_service('blog').config['wp-content']),
            "HelloWorld")

        # TODO verify include-file

        # Verify relations
        self.assertEqual(
            list(d.get_relations()),
            [('blog', 'db'), ('blog', 'cache'), ('blog', 'haproxy')])

    def test_maas_name_and_zone_placement(self):
        d = self.get_named_deployment_v3("stack-placement-maas.yml", "stack")
        d.validate_placement()
        placement = d.get_unit_placement('ceph', {})
        self.assertEqual(placement.get(0), "arnolt")
        placement = d.get_unit_placement('heat', {})
        self.assertEqual(placement.get(0), "zone=zebra")

    @skip_if_offline
    def test_validate_placement_sorting(self):
        d = self.get_named_deployment_and_fetch_v3("stack-placement.yaml", "stack")
        services = d.get_services()
        self.assertEqual(services[0].name, 'nova-compute')
        try:
            d.validate_placement()
        except ErrorExit:
            self.fail("Should not fail")

    def test_machines_placement_sort(self):
        d = Deployment('test', None, None)
        self.assertEqual(
            d._machines_placement_sort(
                FauxService(unit_placement=1),
                FauxService()
            ), 1)
        self.assertEqual(
            d._machines_placement_sort(
                FauxService(),
                FauxService(unit_placement=1)
            ), -1)
        self.assertEqual(
            d._machines_placement_sort(
                FauxService(name="x", unit_placement=['asdf']),
                FauxService(name="y", unit_placement=['lxc:x/1'])
            ), -1)
        self.assertEqual(
            d._machines_placement_sort(
                FauxService(name="y", unit_placement=['lxc:x/1']),
                FauxService(name="x", unit_placement=['asdf'])
            ), 1)
        self.assertEqual(
            d._machines_placement_sort(
                FauxService(name="x", unit_placement=['y']),
                FauxService(name="y")
            ), 1)
        self.assertEqual(
            d._machines_placement_sort(
                FauxService(name="x", unit_placement=['asdf']),
                FauxService(name="y", unit_placement=['hjkl'])
            ), -1)
        self.assertEqual(
            d._machines_placement_sort(
                FauxService(name="x"),
                FauxService(name="y")
            ), -1)

    def test_colocate(self):
        status = {
            'services': {
                'foo': {
                    'units': {
                        '1': {
                            'machine': 1
                        },
                        '2': {}
                    }
                }
            }
        }
        d = self.get_named_deployment_v3("stack-placement.yaml", "stack")
        p = d.get_unit_placement('ceph', status)
        svc = FauxService(name='bar')

        self.assertEqual(p.colocate(status, 'asdf', '1', '', svc),
                         None)
        self.assertIn('Service bar to be deployed with non-existent service '
                      'asdf',
                      self.output.getvalue())
        self.assertEqual(p.colocate(status, 'foo', '2', '', svc),
                         None)
        self.assertIn('Service:bar, Deploy-with-service:foo, '
                      'Requested-unit-index=2, Cannot solve, '
                      'falling back to default placement',
                      self.output.getvalue())
        self.assertEqual(p.colocate(status, 'foo', '1', '', svc),
                         None)
        self.assertIn('Service:bar deploy-with unit missing machine 2',
                      self.output.getvalue())
        self.assertEqual(p.colocate(status, 'foo', '0', '', svc), 1)

    @skip_if_offline
    def test_validate_invalid_placement_nested(self):
        d = self.get_named_deployment_and_fetch_v3("stack-placement-invalid.yaml", "stack")
        services = d.get_services()
        self.assertEqual(services[0].name, 'nova-compute')
        try:
            d.validate_placement()
        except ErrorExit:
            pass
        else:
            self.fail("Should fail")

    @skip_if_offline
    def test_validate_invalid_placement_no_with_service(self):
        d = self.get_named_deployment_and_fetch_v3(
            "stack-placement-invalid-2.yaml", "stack")
        services = d.get_services()
        self.assertEqual(services[0].name, 'nova-compute')
        try:
            d.validate_placement()
        except ErrorExit:
            pass
        else:
            self.fail("Should fail")

    @skip_if_offline
    def test_validate_invalid_placement_subordinate_v3(self):
        # Placement validation fails if a subordinate charm is provided.
        deployment = self.get_named_deployment_and_fetch_v3(
            'stack-placement-invalid-subordinate.yaml', 'stack')
        with self.assertRaises(ErrorExit):
            deployment.validate_placement()
        output = self.output.getvalue()
        self.assertIn(
            'Cannot place to a subordinate service: ceph -> nrpe\n', output)
        self.assertIn(
            'Cannot place to a subordinate service: nova-compute -> nrpe\n',
            output)

    @skip_if_offline
    def test_validate_invalid_placement_subordinate_v4(self):
        # Placement validation fails if a subordinate charm is provided.
        deployment = self.get_deployment_and_fetch_v4(
            'placement-invalid-subordinate.yaml')
        with self.assertRaises(ErrorExit):
            deployment.validate_placement()
        output = self.output.getvalue()
        self.assertIn(
            'Cannot place to a subordinate service: nova-compute -> nrpe\n',
            output)

    def test_validate_invalid_unit_number(self):
        # Placement validation fails if an invalid unit number is provided.
        deployment = self.get_deployment_v4('placement-invalid-number.yaml')
        with self.assertRaises(ErrorExit):
            deployment.validate_placement()
        output = self.output.getvalue()
        self.assertIn(
            'Invalid unit number for placement: django to bad-wolf\n', output)

    def test_get_unit_placement_v3(self):
        d = self.get_named_deployment_v3("stack-placement.yaml", "stack")
        status = {
            'services': {
                'nova-compute': {
                    'units': {
                        'nova-compute/2': {'machine': '1'},
                        'nova-compute/3': {'machine': '2'},
                        'nova-compute/4': {'machine': '3'}}}}}
        placement = d.get_unit_placement('ceph', status)
        self.assertEqual(placement.get(0), '1')
        self.assertEqual(placement.get(1), '2')
        self.assertEqual(placement.get(2), None)

        placement = d.get_unit_placement('quantum', status)
        self.assertEqual(placement.get(0), 'lxc:1')
        self.assertEqual(placement.get(2), 'lxc:3')
        self.assertEqual(placement.get(3), None)

        placement = d.get_unit_placement('verity', status)
        self.assertEqual(placement.get(0), 'lxc:3')

        placement = d.get_unit_placement('mysql', status)
        self.assertEqual(placement.get(0), '0')

        placement = d.get_unit_placement('semper', status)
        self.assertEqual(placement.get(0), '3')

        placement = d.get_unit_placement('lxc-service', status)
        self.assertEqual(placement.get(0), 'lxc:2')
        self.assertEqual(placement.get(1), 'lxc:3')
        self.assertEqual(placement.get(2), 'lxc:1')
        self.assertEqual(placement.get(3), 'lxc:1')
        self.assertEqual(placement.get(4), 'lxc:3')

    def test_fill_placement_v4(self):
        d = self.get_deployment_v4('fill_placement.yaml')
        self.assertEqual(
            d.get_unit_placement('mediawiki1', 0).service.svc_data['to'],
            ['new', 'new'])
        self.assertEqual(
            d.get_unit_placement('mediawiki2', 0).service.svc_data['to'],
            ['0', '0'])
        self.assertEqual(
            d.get_unit_placement('mediawiki3', 0).service.svc_data['to'],
            ['mediawiki1/0', 'mediawiki1/1'])

    def test_parse_placement_v4(self):
        # Short-cut to winding up with a valid placement.
        d = self.get_deployment_v4('simple.yaml')
        placement = d.get_unit_placement('mysql', {})

        c, p, u = placement._parse_placement('mysql')
        self.assertEqual(c, None)
        self.assertEqual(p, 'mysql')
        self.assertEqual(u, None)

        c, p, u = placement._parse_placement('mysql/1')
        self.assertEqual(c, None)
        self.assertEqual(p, 'mysql')
        self.assertEqual(u, '1')

        c, p, u = placement._parse_placement('lxc:mysql')
        self.assertEqual(c, 'lxc')
        self.assertEqual(p, 'mysql')
        self.assertEqual(u, None)

    def test_validate_v4(self):
        d = self.get_deployment_v4('validate.yaml')
        placement = d.get_unit_placement('mysql', {})
        feedback = placement.validate()
        self.assertEqual(feedback.get_errors(), [
            'Invalid container type: asdf service: mysql placement: asdf:0',
            'Service placement to machine not supported: mysql to asdf:0',
            'Invalid service placement: mysql to lxc:asdf',
            'Service placement to machine not supported: mysql to 1',
            'Service unit does not exist: mysql to wordpress/3',
            'Invalid service placement: mysql to asdf'])

    def test_get_unit_placement_v4_simple(self):
        d = self.get_deployment_v4('simple.yaml')
        placement = d.get_unit_placement('mysql', {})
        self.assertEqual(placement.get(0), None)

        placement = d.get_unit_placement('mediawiki', {})
        self.assertEqual(placement.get(0), None)

    def test_get_unit_placement_v4_placement(self):
        d = self.get_deployment_v4('placement.yaml')
        machines = {
            '1': 1,
            '2': 2,
        }

        d.set_machines(machines)

        placement = d.get_unit_placement('mysql', {})
        d.set_machines(machines)
        self.assertEqual(placement.get(0), 2)

        placement = d.get_unit_placement('mediawiki', {})
        self.assertEqual(placement.get(0), 1)

    def test_get_unit_placement_v4_hulk_smash(self):
        d = self.get_deployment_v4('hulk-smash.yaml')
        machines = {
            '1': 1,
        }
        status = {
            'services': {
                'mediawiki': {
                    'units': {
                        'mediawiki/1': {'machine': 1}
                    }
                }
            }
        }

        d.set_machines(machines)

        placement = d.get_unit_placement('mysql', status)
        self.assertEqual(placement.get(0), 1)

        placement = d.get_unit_placement('mediawiki', status)
        self.assertEqual(placement.get(0), 1)

    def test_get_unit_placement_v4_hulk_smash_nounits(self):
        d = self.get_deployment_v4('hulk-smash-nounits.yaml')
        machines = {
            '1': 1,
        }
        status = {
            'services': {
                'mediawiki': {
                    'units': {
                        'mediawiki/1': {'machine': 1}
                    }
                }
            }
        }

        d.set_machines(machines)

        placement = d.get_unit_placement('mysql', status)
        self.assertEqual(placement.get(0), 1)

        placement = d.get_unit_placement('mediawiki', status)
        self.assertEqual(placement.get(0), 1)

    def test_get_unit_placement_v4_hulk_smash_nounits_nomachines(self):
        d = self.get_deployment_v4('hulk-smash-nounits-nomachines.yaml')
        machines = {
            '1': 1,
        }
        status = {
            'services': {
                'mediawiki': {
                    'units': {
                        'mediawiki/1': {'machine': 1}
                    }
                }
            }
        }

        d.set_machines(machines)

        placement = d.get_unit_placement('mysql', status)
        self.assertEqual(placement.get(0), 1)

        # Since we don't have a placement, even with the status, this should
        # still be None.
        placement = d.get_unit_placement('mediawiki', status)
        self.assertEqual(placement.get(0), None)

    def test_get_unit_placement_v4_container(self):
        d = self.get_deployment_v4('container.yaml')
        machines = {
            '1': 1,
        }
        status = {
            'services': {
                'mediawiki': {
                    'units': {
                        'mediawiki/1': {'machine': 1},
                    }
                }
            }
        }

        d.set_machines(machines)

        placement = d.get_unit_placement('mysql', status)
        self.assertEqual(placement.get(0), 'lxc:1')

        placement = d.get_unit_placement('mediawiki', status)
        self.assertEqual(placement.get(0), 1)

    def test_get_unit_placement_v4_container_new(self):
        d = self.get_deployment_v4('container-new.yaml')
        machines = {
            '1': 1,
            'mysql/0': 2
        }
        status = {
            'services': {
                'mediawiki': {
                    'units': {
                        'mediawiki/1': {'machine': 1}
                    }
                }
            }
        }

        d.set_machines(machines)

        placement = d.get_unit_placement('mysql', status)
        self.assertEqual(placement.get_new_machines_for_containers(),
            ['mysql/0'])
        self.assertEqual(placement.get(0), 'lxc:2')

        placement = d.get_unit_placement('mediawiki', status)
        self.assertEqual(placement.get(0), 1)

    def test_multiple_relations_no_weight(self):
        data = {"relations": {"wordpress": {"consumes": ["mysql"]},
                              "nginx": {"consumes": ["wordpress"]}}}
        d = Deployment("foo", data, include_dirs=())
        self.assertEqual(
            [('nginx', 'wordpress'), ('wordpress', 'mysql')],
            list(d.get_relations()))

    def test_multiple_relations_weighted(self):
        data = {
            "relations": {
                "keystone": {
                    "weight": 100,
                    "consumes": ["mysql"]
                },
                "nova-compute": {
                    "weight": 50,
                    "consumes": ["mysql"]
                },
                "glance": {
                    "weight": 70,
                    "consumes": ["mysql"]
                },
            }
        }
        d = Deployment("foo", data, include_dirs=())
        self.assertEqual(
            [('keystone', 'mysql'), ('glance', 'mysql'),
             ('nova-compute', 'mysql')], list(d.get_relations()))

    def test_getting_service_names(self):
        # It is possible to retrieve the service names.
        deployment = self.get_named_deployment_v3("stack-placement.yaml", "stack")
        service_names = deployment.get_service_names()
        expected_service_names = [
            'ceph', 'mysql', 'nova-compute', 'quantum', 'semper', 'verity', 'lxc-service']
        self.assertEqual(set(expected_service_names), set(service_names))

    def test_resolve_config_handles_empty_options(self):
        """resolve_config should handle options being "empty" lp:1361883"""
        deployment = self.get_named_deployment_v3("negative.cfg", "negative")
        self.assertEqual(
            deployment.data["services"]["foo"]["options"], {})
        deployment.resolve_config()

    def test_resolve_config_handles_none_options(self):
        """resolve_config should handle options being "none" lp:1361883"""
        deployment = self.get_named_deployment_v3("negative.yaml", "negative")
        self.assertEqual(
            deployment.data["services"]["foo"]["options"], None)
        deployment.resolve_config()



import logging
import os
import urllib2
import shutil

from .vcs import Git, Bzr
from .utils import (
    _check_call,
    _get_juju_home,
    extract_zip,
    get_qualified_charm_url,
    path_join,
    path_exists,
    STORE_URL,
    temp_file,
    yaml_load)


class Charm(object):

    log = logging.getLogger('deployer.charm')

    def __init__(self, name, path, branch, rev, build, charm_url=""):
        self.name = name
        self._path = path
        self.branch = branch
        self.rev = rev
        self._charm_url = charm_url
        self._build = build
        self.vcs = self.get_vcs()

    def get_vcs(self):
        if not self.branch:
            return None
        if self.branch.startswith('git') or 'github.com' in self.branch or \
                'git.launchpad.net' in self.branch or \
                os.path.exists(os.path.join(self.branch, '.git')):
            return Git(self.path, self.branch, self.log)
        elif self.branch.startswith("bzr") or self.branch.startswith('lp:') \
                or os.path.exists(os.path.join(self.branch, '.bzr')) \
                or self.branch.startswith('file:'):
            return Bzr(self.path, self.branch, self.log)
        raise ValueError(
            "Could not determine vcs backend for %s" % self.branch)

    @classmethod
    def from_service(cls, name, repo_path, deploy_series, data):
        """
        name: service name
        data['charm']: charm name or store charm url
        data['charm_url'] store charm url
        """
        branch, rev, series = None, None, None
        charm_branch = data.get('branch')
        if charm_branch is not None:
            branch, sep, rev = charm_branch.partition('@')

        charm_path, store_url, build = None, None, None
        name = data.get('charm', name)
        if name.startswith('cs:'):
            store_url = name
        elif name.startswith('local:'):
            # Support vcs charms specifying their
            parts = name[len('local:'):].split('/')
            if len(parts) == 2:
                series, name = parts
            elif data.get('series'):
                series = data['series']
                name = parts.pop()
            else:
                series = deploy_series
            charm_path = path_join(repo_path, series, name)
        elif os.path.isabs(name):
            # charm points to an absolute local path
            charm_path = name.rstrip(os.path.sep)
        elif 'series' in data:
            series = data['series']
            charm_path = path_join(repo_path, series, name)
        else:
            charm_path = path_join(repo_path, deploy_series, name)

        if not store_url:
            store_url = data.get('charm_url', None)

        if store_url and branch:
            cls.log.error(
                "Service: %s has both charm url: %s and branch: %s specified",
                name, store_url, branch)
        if not store_url and not branch and not os.path.isabs(name):
            cls.log.error(
                "Service: %s has neither charm url or branch specified",
                name)
        elif not store_url:
            build = data.get('build', '')

        return cls(name, charm_path, branch, rev, build, store_url)

    def is_absolute(self):
        """Charm config points to an absolute path on disk.

        """
        return os.path.isabs(self.name)

    @property
    def repo_path(self):
        """The Juju repository path in which this charm resides.

        For most charms this returns None, leaving the repo path to be
        determined by the Deployment that deploys the charm.

        For charms at an absolute path, however, the repo path is by
        definition the directory two levels up from the charm. (And the
        series directory is one level up.) This allows us to deploy
        charms from anywhere on the filesystem without first gathering them
        under one repository path.

        """
        if self.is_absolute():
            d = os.path.dirname
            return d(d(self.path))
        return None

    def is_local(self):
        if self._charm_url:
            if self._charm_url.startswith('cs:'):
                return False
        return True

    def exists(self):
        return self.is_local() and path_exists(self.path)

    def is_subordinate(self):
        return self.metadata.get('subordinate', False)

    @property
    def charm_url(self):
        if self._charm_url:
            return self._charm_url
        series = os.path.basename(os.path.dirname(self.path))
        charm_name = self.metadata['name']
        return "local:%s/%s" % (series, charm_name)

    def build(self):
        if not self._build:
            return
        self.log.debug("Building charm %s with %s", self.path, self._build)
        _check_call([self._build], self.log,
                    "Charm build failed %s @ %s", self._build, self.path,
                    cwd=self.path, shell=True)

    def fetch(self):
        if self._charm_url:
            self._fetch_store_charm()
            return
        elif self.is_absolute():
            return
        elif not self.branch:
            self.log.warning("Invalid charm specification %s", self.name)
            return
        self.log.debug(" Branching charm %s @ %s", self.branch, self.path)
        self.vcs.branch()
        if self.rev:
            self.vcs.update(self.rev)
        self.build()

    @property
    def path(self):
        if not self.is_local() and not self._path:
            self._path = self._get_charm_store_cache()
        return self._path

    @property
    def series_path(self):
        if self.is_absolute() or not self.is_local():
            return None
        return os.path.dirname(self.path)

    def _fetch_store_charm(self, update=False):
        cache_dir = self._get_charm_store_cache()
        self.log.debug("Cache dir %s", cache_dir)

        if os.path.exists(cache_dir) and not update:
            return

        qualified_url = get_qualified_charm_url(self.charm_url)

        self.log.debug("Retrieving store charm %s" % qualified_url)

        if update and os.path.exists(cache_dir):
            shutil.rmtree(cache_dir)

        store_url = "%s/charm/%s" % (STORE_URL, qualified_url[3:])
        with temp_file() as fh:
            ufh = urllib2.urlopen(store_url)
            shutil.copyfileobj(ufh, fh)
            fh.flush()
            extract_zip(fh.name, self.path)
        self.config

    def _get_charm_store_cache(self):
        assert not self.is_local(), "Attempt to get store charm for local"
        # Cache
        jhome = _get_juju_home()
        cache_dir = os.path.join(jhome, ".deployer-store-cache")
        if not os.path.exists(cache_dir):
            os.mkdir(cache_dir)
        return os.path.join(
            cache_dir,
            self.charm_url.replace(':', '_').replace("/", "_"))

    def update(self, build=False):
        if not self.branch:
            return
        assert self.exists()
        self.log.debug(" Updating charm %s from %s", self.path, self.branch)
        self.vcs.update(self.rev)
        if build:
            self.build()

    def is_modified(self):
        if not self.branch:
            return False
        return self.vcs.is_modified()

    @property
    def config(self):
        config_path = path_join(self.path, "config.yaml")
        if not path_exists(config_path):
            return {}

        with open(config_path) as fh:
            return yaml_load(fh.read()).get('options', {})

    @property
    def metadata(self):
        md_path = path_join(self.path, "metadata.yaml")
        if not path_exists(md_path):
            if not path_exists(self.path):
                raise RuntimeError("No charm metadata @ %s", md_path)
        with open(md_path) as fh:
            return yaml_load(fh.read())

    def get_provides(self):
        p = {'juju-info': [{'name': 'juju-info'}]}
        for key, value in self.metadata['provides'].items():
            value['name'] = key
            p.setdefault(value['interface'], []).append(value)
        return p

    def get_requires(self):
        r = {}
        for key, value in self.metadata['requires'].items():
            value['name'] = key
            r.setdefault(value['interface'], []).append(value)
        return r



import inspect
import logging
import os
import unittest
import shutil
import StringIO
import tempfile

import mock

import deployer
from deployer.config import ConfigStack


# Skip during launchpad recipe package builds (DEB_BUILD_ARCH) or if explicitly
# requested with 'TEST_OFFLINE=1'
TEST_OFFLINE = ("DEB_BUILD_ARCH" in os.environ or "TEST_OFFLINE" in os.environ)
TEST_OFFLINE_REASON = "Requires configured bzr launchpad id and network access"
skip_if_offline = unittest.skipIf(TEST_OFFLINE, TEST_OFFLINE_REASON)


class Base(unittest.TestCase):

    test_data_dir = os.path.join(
        os.path.dirname(inspect.getabsfile(deployer)), "tests", "test_data")

    @classmethod
    def setUpClass(cls):
        os.environ["JUJU_HOME"] = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(os.environ["JUJU_HOME"])

    def get_named_deployment_v3(self, file_name, stack_name):
        """ Get v3 deployment from a test_data file.
        """
        return ConfigStack(
            [os.path.join(
                self.test_data_dir, file_name)]).get(stack_name)

    def get_deployment_v4(self, file_name):
        """Get v4 deployment from a test_data file.
        """
        f = os.path.join(self.test_data_dir, 'v4', file_name)
        return ConfigStack([f]).get(f)

    def capture_logging(self, name="", level=logging.INFO,
                        log_file=None, formatter=None):
        if log_file is None:
            log_file = StringIO.StringIO()
        log_handler = logging.StreamHandler(log_file)
        if formatter:
            log_handler.setFormatter(formatter)
        logger = logging.getLogger(name)
        logger.addHandler(log_handler)
        old_logger_level = logger.level
        logger.setLevel(level)

        @self.addCleanup
        def reset_logging():
            logger.removeHandler(log_handler)
            logger.setLevel(old_logger_level)
        return log_file

    def mkdir(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        return d

    def change_environment(self, **kw):
        """
        """
        original_environ = dict(os.environ)

        @self.addCleanup
        def cleanup_env():
            os.environ.clear()
            os.environ.update(original_environ)

        os.environ.update(kw)


def patch_env_status(env, services, machines=()):
    """Simulate that the given mock env has the status described in services.

    This function is used so that tests do not have to wait minutes for
    service units presence when the importer is used with the given env.

    The services argument is a dict mapping service names with the number of
    their units. This will be reflected by the status returned when the
    importer adds the units (see "deployer.action.importer.Importer.add_unit").

    The machines argument can be used to simulate that the given machines are
    present in the Juju environment.
    """
    services_status = dict(
        (k, {'units': dict((i, {}) for i in range(v))})
        for k, v in services.items()
    )
    machines_status = dict((i, {}) for i in machines)
    env.status.side_effect = [
        # There are no services initially.
        {'services': {}, 'machines': machines_status},
        {'services': {}, 'machines': machines_status},
        # This is the workaround check for subordinate charms presence:
        # see lp:1421315 for details.
        {'services': services_status, 'machines': machines_status},
        {'services': services_status, 'machines': machines_status},
        # After we exited the workaround loop, we can just mock further
        # status results.
        mock.MagicMock(),
        mock.MagicMock(),
        mock.MagicMock(),
    ]



from base64 import b64encode

import logging
import pprint
import os
import yaml

from .charm import Charm
from .feedback import Feedback
from .service import Service, ServiceUnitPlacementV3, ServiceUnitPlacementV4
from .relation import Endpoint
from .utils import path_join, yaml_dump, ErrorExit, resolve_include, x_in_y


class Deployment(object):

    log = logging.getLogger("deployer.deploy")

    def __init__(self, name, data, include_dirs, repo_path="", version=3):
        self.name = name
        self.data = data
        self.include_dirs = include_dirs
        self.repo_path = repo_path
        self.version = version
        self.machines = {}

    @property
    def series(self):
        # Series could use a little help, charm series should be inferred
        # directly from a store url
        return self.data.get('series', 'precise')

    @property
    def series_path(self):
        return path_join(self.repo_path, self.series)

    def pretty_print(self):
        pprint.pprint(self.data)

    def get_service(self, name):
        if name not in self.data['services']:
            return
        return Service(name, self.data['services'][name])

    def get_services(self):
        services = []
        for name, svc_data in self.data.get('services', {}).items():
            services.append(Service(name, svc_data))
        if self.version == 3:
            # Sort unplaced units first, then sort by name for placed units.
            services.sort(key=lambda svc: (bool(svc.unit_placement), svc.name))
        else:
            services.sort(self._machines_placement_sort)
        return services

    def set_machines(self, machines):
        """Set a dict of machines, mapping from the names in the machine spec
        to the machine names in the environment status.
        """
        self.machines = machines

    def get_machine(self, id):
        """Return a dict containing machine options
        None is returned if the machine doesn't exists in
        the bundle YAML."""
        return self.get_machines().get(id, None)

    def get_machines(self):
        """Return a dict mapping machine names to machine options.

        An empty dict is returned if no machines are defined in the
        bundle YAML.
        """
        machines = {}
        for key, machine in self.data.get('machines', {}).items():
            machines[str(key)] = machine
        return machines

    def get_service_names(self):
        """Return a sequence of service names for this deployment."""
        return self.data.get('services', {}).keys()

    @staticmethod
    def _machines_placement_sort(svc_a, svc_b):
        """Sort machines with machine placement in mind.

        If svc_a is colocated alongside svc_b, svc_b needs to be deployed
        first, so that it exists by the time svc_a is deployed, and vice
        versa; this sorts first based on this fact, then secondly based on
        whether or not the service has a unit placement, and then finally
        based on the name of the service.
        """
        if svc_a.unit_placement:
            if svc_b.unit_placement:
                # Check for colocation.  This naively assumes that there is no
                # circularity in placements.
                if x_in_y(svc_b, svc_a):
                    return 1
                if x_in_y(svc_a, svc_b):
                    return -1
                # If no colocation exists, simply compare names.
                return cmp(svc_a.name, svc_b.name)
            return 1
        if svc_b.unit_placement:
            return -1
        return cmp(svc_a.name, svc_b.name)

    def get_unit_placement(self, svc, status):
        if isinstance(svc, (str, unicode)):
            svc = self.get_service(svc)
        if self.version == 3:
            return ServiceUnitPlacementV3(svc, self, status)
        else:
            return ServiceUnitPlacementV4(svc, self, status,
                                          machines_map=self.machines)

    def get_relations(self):
        if 'relations' not in self.data:
            return

        # Strip duplicate rels
        seen = set()

        def check(a, b):
            k = tuple(sorted([a, b]))
            if k in seen:
                #self.log.warning(" Skipping duplicate relation %r" % (k,))
                return
            seen.add(k)
            return True

        # Support an ordered list of [endpoints]
        if isinstance(self.data['relations'], list):
            for end_a, end_b in self.data['relations']:
                # Allow shorthand of [end_a, [end_b, end_c]]
                if isinstance(end_b, list):
                    for eb in end_b:
                        if check(end_a, eb):
                            yield (end_a, eb)
                else:
                    if check(end_a, end_b):
                        yield (end_a, end_b)
            return

        # Legacy format (dictionary of dictionaries with weights)
        rels = {}
        for k, v in self.data['relations'].items():
            expanded = []
            for c in v['consumes']:
                expanded.append((k, c))
            by_weight = rels.setdefault(v.get('weight', 0), [])
            by_weight.extend(expanded)
        for k in sorted(rels, reverse=True):
            for r in rels[k]:
                if check(*r):
                    yield r
        #self.log.debug(
        #    "Found relations %s\n  %s" % (" ".join(map(str, seen))))

    def get_charms(self):
        for k, v in self.data.get('services', {}).items():
            yield Charm.from_service(k, self.repo_path, self.series, v)

    def get_charm_for(self, svc_name):
        svc_data = self.data['services'][svc_name]
        return Charm.from_service(
            svc_name, self.repo_path, self.series, svc_data)

    def fetch_charms(self, update=False, no_local_mods=False):
        for charm in self.get_charms():
            if charm.is_local():
                if charm.exists():
                    if no_local_mods and charm.is_modified():
                        self.log.warning(
                            "Charm %r has local modifications",
                            charm.path)
                        raise ErrorExit()
                    if charm.rev or update:
                        if charm.rev and charm.is_modified():
                            self.log.warning((
                                "Charm %r with pinned revision has"
                                " local modifications"), charm.path)
                            raise ErrorExit()
                        charm.update(build=True)
                    continue
                elif not os.path.exists(charm.series_path):
                    os.mkdir(charm.series_path)
            charm.fetch()

    def resolve(self, cli_overides=()):
        # Once we have charms definitions available, we can do verification
        # of config options.
        self.load_overrides(cli_overides)
        self.resolve_config()
        self.validate_relations()
        self.validate_placement()

    def load_overrides(self, cli_overrides=()):
        """Load overrides."""
        overrides = {}
        overrides.update(self.data.get('overrides', {}))

        for o in cli_overrides:
            key, value = o.split('=', 1)
            overrides[key] = value

        for k, v in overrides.iteritems():
            found = False
            for svc_name, svc_data in self.data['services'].items():
                charm = self.get_charm_for(svc_name)
                if k in charm.config:
                    if not svc_data.get('options'):
                        svc_data['options'] = {}
                    svc_data['options'][k] = v
                    found = True
            if not found:
                self.log.warning(
                    "Override %s does not match any charms", k)

    def resolve_config(self):
        """Load any lazy config values (includes), and verify config options.
        """
        self.log.debug("Resolving configuration")
        # XXX TODO, rename resolve, validate relations
        # against defined services
        feedback = Feedback()
        for svc_name, svc_data in self.data.get('services', {}).items():
            if not 'options' in svc_data:
                continue
            charm = self.get_charm_for(svc_name)
            config = charm.config
            options = {}

            svc_options = svc_data.get('options', {})
            if svc_options is None:
                svc_options = {}
            for k, v in svc_options.items():
                if not k in config:
                    feedback.error(
                        "Invalid config charm %s %s=%s" % (charm.name, k, v))
                    continue
                iv = self._resolve_include(svc_name, k, v)
                if isinstance(iv, Feedback):
                    feedback.extend(iv)
                    continue
                if iv is not None:
                    v = iv
                options[k] = v
            svc_data['options'] = options
        self._handle_feedback(feedback)

    def _resolve_include(self, svc_name, k, v):
        feedback = Feedback()
        for include_type in ["file", "base64"]:
            if (not isinstance(v, basestring)
                or not v.startswith(
                    "include-%s://" % include_type)):
                continue
            include, fname = v.split("://", 1)
            ip = resolve_include(fname, self.include_dirs)
            if ip is None:
                feedback.error(
                    "Invalid config %s.%s include not found %s" % (
                    svc_name, k, v))
                continue
            with open(ip) as fh:
                v = fh.read()
                if include_type == "base64":
                    v = b64encode(v)
                return v
        if feedback:
            return feedback

    def validate_relations(self):
        # Could extend to do interface matching against charms.
        services = dict([(s.name, s) for s in self.get_services()])
        feedback = Feedback()
        for e_a, e_b in self.get_relations():
            for ep in [Endpoint(e_a), Endpoint(e_b)]:
                if not ep.service in services:
                    feedback.error(
                        ("Invalid relation in config,"
                         " service %s not found, rel %s") % (
                             ep.service, "%s <-> %s" % (e_a, e_b)))
                    continue
        self._handle_feedback(feedback)

    def validate_placement(self):
        services = dict([(s.name, s) for s in self.get_services()])
        feedback = Feedback()
        for name, s in services.items():
            placement = self.get_unit_placement(s, {})
            feedback.extend(placement.validate())
        self._handle_feedback(feedback)

    def _handle_feedback(self, feedback):
        for e in feedback.get_errors():
            self.log.error(e)
        for w in feedback.get_warnings():
            self.log.warning(w)
        if feedback.has_errors:
            raise ErrorExit()

    def save(self, path):
        with open(path, "w") as fh:
            fh.write(yaml_dump(self.data))

    @staticmethod
    def to_yaml(dumper, deployment):
        return dumper.represent_dict(deployment.data)

yaml.add_representer(Deployment, Deployment.to_yaml)



"""Juju GUI server bundles deployment support.

The following functions are used by the Juju GUI server to validate
and start bundle deployments. The validate and import_bundle
operations represents the public API: they are directly called in the
GUI server bundles support code, which also takes care of handling any
exception they can raise.  Those functions are blocking, and therefore
the GUI server executes them in separate processes.  See
<https://code.launchpad.net/~juju-gui/charms/precise/juju-gui/trunk>.
"""

import os

from deployer.action.importer import Importer
from deployer.cli import setup_parser
from deployer.deployment import Deployment
from deployer.env.gui import GUIEnvironment
from deployer.utils import (
    DeploymentError,
    mkdir,
)


# This value is used by the juju-deployer Importer object to store charms.
# This directory is usually created in the machine where the Juju GUI charm is
# deployed the first time a bundle deployment is requested.
JUJU_HOME = '/var/lib/juju-gui/juju-home'


def get_default_guiserver_options():
    """Return the default importer options used by the GUI server."""
    # Options used by the juju-deployer.  The defaults work for us, except for
    # the ignore_errors flag.
    return setup_parser().parse_args(['--ignore-errors'])


class GUIDeployment(Deployment):
    """Handle bundle deployments requested by the GUI server."""

    def __init__(self, name, data):
        super(GUIDeployment, self).__init__(name, data, [])

    def _handle_feedback(self, feedback):
        """Raise a DeploymentError if the given feedback includes errors.

        The GUI server will catch and report failures propagating them through
        the WebSocket connection to the client.
        """
        for message in feedback.get_warnings():
            self.log.warning(message)
        if feedback.has_errors:
            # Errors are logged by the GUI server.
            raise DeploymentError(feedback.get_errors())


def _validate(env, bundle):
    """Bundle validation logic, used by both validate and import_bundle.

    This function receives a connected environment and the bundle as a YAML
    decoded object.
    """
    # Retrieve the services deployed in the Juju environment.
    env_status = env.status()
    env_services = set(env_status['services'].keys())
    # Retrieve the services in the bundle.
    bundle_services = set(bundle.get('services', {}).keys())
    # Calculate overlapping services.
    overlapping = env_services.intersection(bundle_services)
    if overlapping:
        services = ', '.join(overlapping)
        error = 'service(s) already in the environment: {}'.format(services)
        raise ValueError(error)


def validate(apiurl, username, password, bundle):
    """Validate a bundle."""
    env = GUIEnvironment(apiurl, username, password)
    env.connect()
    try:
        _validate(env, bundle)
    finally:
        env.close()


def import_bundle(apiurl, username, password, name, bundle, options):
    """Import a bundle."""
    env = GUIEnvironment(apiurl, username, password)
    deployment = GUIDeployment(name, bundle)
    importer = Importer(env, deployment, options)
    env.connect()
    # The Importer tries to retrieve the Juju home from the JUJU_HOME
    # environment variable: create a customized directory (if required) and
    # set up the environment context for the Importer.
    mkdir(JUJU_HOME)
    os.environ['JUJU_HOME'] = JUJU_HOME
    try:
        _validate(env, bundle)
        importer.run()
    finally:
        env.close()



from deployer.utils import parse_constraints
from jujuclient import (UnitErrors,
                        EnvError)


class MemoryEnvironment(object):
    """ In memory deployment: not all features implemented
        (notably subordinates and their relations).
    """

    def __init__(self, name, deployment):
        """ Two main dicts: _services (return-able as part of status(),
            and _services_data (to hold e.g. config, constraints)
        """
        super(MemoryEnvironment, self).__init__()
        self.name = name
        self._deployment = deployment
        self._services = {}
        self._services_data = {}
        self._relations = {}
        self._do_deploy()

    def add_units(self, svc_name, num):
        """Add units
        """
        next_num = self._services_data[svc_name]['next_unit_num']
        for idx in xrange(next_num, next_num + num):
            self._services[svc_name]['units'].append(
                '{}/{}'.format(svc_name, idx))
        self._services_data[svc_name]['next_unit_num'] = \
            next_num + num

    def remove_unit(self, unit_name):
        """ Remove a unit by name """
        svc_name = unit_name.split('/')[0]
        units_idx = {unit: idx for idx, unit in
                     enumerate(self._services[svc_name]['units'])}
        try:
            self._services[svc_name]['units'].pop(
                units_idx[unit_name])
        except KeyError:
            raise UnitErrors("Invalid unit name")

    def _get_service(self, svc_name):
        """ Get service by name (as returned by status())
        """
        if not svc_name in self._services:
            raise EnvError("Invalid service name")
        return self._services[svc_name]

    def add_relation(self, endpoint_a, endpoint_b):
        """Add relations
        """

    def destroy_service(self, svc_name):
        """ Destroy a service
        """
        if not svc_name in self._services:
            raise EnvError("Invalid service name")
        del self._services[svc_name]

    def close(self):
        """
        """

    def connect(self):
        """
        """

    def set_config(self, svc_name, cfg_dict):
        """ Set service config from passed dict, keeping
            the structure as needed for status() return
        """
        config = self.get_config(svc_name)
        if cfg_dict:
            for cfg_k, cfg_v in cfg_dict.items():
                config_entry = config.setdefault(cfg_k, {})
                config_entry['value'] = cfg_v

    def set_constraints(self, svc_name, constr_str):
        """ Set service constraints from "key=value ..."
            passed string
        """
        constraints = parse_constraints(constr_str) if constr_str else {}
        self._services_data[svc_name]['constraints'] = constraints

    def get_config(self, svc_name):
        """ Return service configs - note its structure:
            config{thename: {'value': thevalue}, ...}
        """
        return self._services_data[svc_name]['config']

    def get_constraints(self, svc_name):
        """ Return service constraints dict
        """
        return self._services_data[svc_name]['constraints']

    def get_cli_status(self):
        pass

    def reset(self):
        pass

    def resolve_errors(self, retry_count=0, timeout=600, watch=False, delay=5):
        pass

    def _do_deploy(self):
        """ Fake deploy: build in-memory representation of the deployed set
            of services from deployment
        """
        self._compile_relations()
        for service in self._deployment.get_services():
            svc_name = service.name
            charm = self._deployment.get_charm_for(svc_name)
            relations = self._relations.setdefault(svc_name, {})
            self._services[svc_name] = {
                'units': [],
                'charm': charm.name,
                'relations': relations,
            }
            self._services_data[svc_name] = {
                'next_unit_num': 0,
                'config': {},
                'constraints': {},
            }
            # XXX: Incomplete relations support: only add units for non-subords
            num_units = 0 if charm.is_subordinate() else service.num_units
            self.add_units(svc_name, num_units)
            self.set_config(svc_name, service.config)
            self.set_constraints(svc_name, service.constraints)

    def _compile_relations(self):
        """ Compile a relation dictionary by svc_name, with
            their values structured for status() return
        """
        for rel in self._deployment.get_relations():
            for src, dst in (rel[0], rel[1]), (rel[1], rel[0]):
                try:
                    src_requires = self._deployment.get_charm_for(
                        src).get_requires()
                    dst_provides = self._deployment.get_charm_for(
                        dst).get_provides()
                except KeyError:
                    continue
                # Create dicts key-ed by:
                #  { interface_type: (interface_name, svc_name)}
                src_dict = {}
                dst_dict = {}
                # {rel_name: [{interface: name }...]}
                for _, interfaces in src_requires.items():
                    for interface in interfaces:
                        src_dict[interface.get('interface')] = (
                            interface.get('name'), src)
                for _, interfaces in dst_provides.items():
                    for interface in interfaces:
                        dst_dict[interface.get('interface')] = (
                            interface.get('name'), dst)
                # Create juju env relation entries as:
                # {svc_name: { interface_name: [ svc_name2, ...] }, ...}
                for src_rel, (if_name, src_svc_name) in src_dict.items():
                    if src_rel in dst_dict:
                        src_rels = self._relations.setdefault(src_svc_name, {})
                        src_rels.setdefault(if_name, [])
                        dst_svc_name = dst_dict[src_rel][1]
                        src_rels[if_name].append(dst_svc_name)

    def status(self):
        """ Return all services status """
        return {'services': self._services}

    def wait_for_units(self, *args, **kwargs):
        pass
